/* All RGB2YUV conversion code and some other parts of code has been taken from testoverlay.c */

/* NOTE: These RGB conversion functions are not intended for speed,
         only as examples.
*/

#include "yuvconv.h"

void RGBtoYUV(Uint8 *rgb, int *yuv, int monochrome, int luminance)
{
    int i;
    Uint8 r;
    Uint8 g;
    Uint8 b;
    Uint16 col;

    col=*(Uint16 *)rgb;
    /* Using RGB565 */
    r=(col&0xf800)>>8;
    g=(col&0x07e0)>>3;
    b=(col&0x001f)<<3;
    if (monochrome)
    {
#if 0 /* these are the two formulas that I found on the FourCC site... */
        yuv[0] = 0.299*r + 0.587*g + 0.114*b;
        yuv[1] = 128;
        yuv[2] = 128;
#else
        yuv[0] = (0.257 * r) + (0.504 * g) + (0.098 * b) + 16;
        yuv[1] = 128;
        yuv[2] = 128;
#endif
    }
    else
    {
#if 0 /* these are the two formulas that I found on the FourCC site... */
        yuv[0] = 0.299*r + 0.587*g + 0.114*b;
        yuv[1] = (b-yuv[0])*0.565 + 128;
        yuv[2] = (r-yuv[0])*0.713 + 128;
#else
        yuv[0] = (0.257 * r) + (0.504 * g) + (0.098 * b) + 16;
        yuv[1] = 128 - (0.148 * r) - (0.291 * g) + (0.439 * b);
        yuv[2] = 128 + (0.439 * r) - (0.368 * g) - (0.071 * b);
#endif
    }

    if (luminance!=100)
    {
        yuv[0]=yuv[0]*luminance/100;
        if (yuv[0]>255)
            yuv[0]=255;
    }
}

void ConvertRGBtoYUY2(SDL_Surface *s, SDL_Overlay *o, int monochrome, int luminance)
{
	int x,y;
	int yuv[3];
	Uint8 *p,*op;

	sdl.LockSurface(s);
	sdl.LockYUVOverlay(o);

	for(y=0; y<s->h && y<o->h; y++)
	{
		p=((Uint8 *) s->pixels)+s->pitch*y;
		op=o->pixels[0]+o->pitches[0]*y;
		for(x=0; x<s->w && x<o->w; x++)
		{
			RGBtoYUV(p,yuv, monochrome, luminance);
			if(x%2==0)
			{
				*(op++)=yuv[0];
				*(op++)=yuv[1];
				op[1]=yuv[2];
			}
			else
			{
				*op=yuv[0];
				op+=2;
			}

			p+=s->format->BytesPerPixel;
		}
	}

	sdl.UnlockYUVOverlay(o);
	sdl.UnlockSurface(s);
}
